// file      : bpkg/pkg-purge.cxx -*- C++ -*-
// copyright : Copyright (c) 2014-2016 Code Synthesis Ltd
// license   : MIT; see accompanying LICENSE file

#include <bpkg/pkg-purge>

#include <bpkg/package>
#include <bpkg/package-odb>
#include <bpkg/database>
#include <bpkg/diagnostics>

using namespace std;
using namespace butl;

namespace bpkg
{
  void
  pkg_purge_fs (const dir_path& c,
                transaction& t,
                const shared_ptr<selected_package>& p,
                bool archive)
  {
    tracer trace ("pkg_purge_archive");

    assert (p->state == package_state::fetched ||
            p->state == package_state::unpacked);

    database& db (t.database ());
    tracer_guard tg (db, trace);

    try
    {
      if (p->purge_src)
      {
        dir_path d (p->src_root->absolute () ? *p->src_root : c / *p->src_root);

        if (exists (d)) // Don't complain if someone did our job for us.
          rm_r (d);

        p->src_root = nullopt;
        p->purge_src = false;
      }

      if (p->purge_archive && archive)
      {
        path a (p->archive->absolute () ? *p->archive : c / *p->archive);

        if (exists (a))
          rm (a);

        p->archive = nullopt;
        p->purge_archive = false;
      }
    }
    catch (const failed&)
    {
      p->state = package_state::broken;
      db.update (p);
      t.commit ();

      info << "package " << p->name << " is now broken; "
           << "use 'pkg-purge --force' to remove";
      throw;
    }
  }

  void
  pkg_purge (const dir_path& c,
             transaction& t,
             const shared_ptr<selected_package>& p)
  {
    assert (p->state == package_state::fetched ||
            p->state == package_state::unpacked);

    tracer trace ("pkg_purge");

    database& db (t.database ());
    tracer_guard tg (db, trace);

    assert (!p->out_root);
    pkg_purge_fs (c, t, p, true);

    db.erase (p);
    t.commit ();

    p->state = package_state::transient;
  }

  int
  pkg_purge (const pkg_purge_options& o, cli::scanner& args)
  {
    tracer trace ("pkg_purge");

    const dir_path& c (o.directory ());
    l4 ([&]{trace << "configuration: " << c;});

    if (!args.more ())
      fail << "package name argument expected" <<
        info << "run 'bpkg help pkg-purge' for more information";

    string n (args.next ());

    database db (open (c, trace));
    transaction t (db.begin ());

    shared_ptr<selected_package> p (db.find<selected_package> (n));

    if (p == nullptr)
      fail << "package " << n << " does not exist in configuration " << c;

    // Make sure the package is in a state from which it can be purged.
    //
    switch (p->state)
    {
    case package_state::fetched:
      {
        // If we have --keep, then this is a no-op. We could have
        // detected this and returned but we still want the normal
        // diagnostics. So instead the logic below takes care of
        // this situation.
        //
        break;
      }
    case package_state::unpacked:
      {
        if (o.keep () && !p->archive)
          fail << "package " << n << " has no archive to keep";

        break;
      }
    case package_state::broken:
      {
        if (!o.force ())
          fail << "broken package " << n << " can only be purged with --force";

        if (o.keep ())
          fail << "cannot keep broken package " << n;

        break;
      }
    default:
      {
        fail << p->state << " package " << n << " cannot be purged";
      }
    }

    // For a broken package we just verify that all the filesystem objects
    // were cleaned up by the user.
    //
    if (p->state == package_state::broken)
    {
      if (p->out_root)
      {
        dir_path d (c / *p->out_root); // Always relative.

        if (exists (d))
          fail << "output directory of broken package " << n
               << " still exists" <<
            info << "remove " << d << " manually then re-run pkg-purge";
      }

      if (p->purge_src)
      {
        dir_path d (p->src_root->absolute () ? *p->src_root : c / *p->src_root);

        if (exists (d))
          fail << "source directory of broken package " << n
               << " still exists" <<
            info << "remove " << d << " manually then re-run pkg-purge";
      }

      if (p->purge_archive)
      {
        path a (p->archive->absolute () ? *p->archive : c / *p->archive);

        if (exists (a))
          fail << "archive file of broken package " << n << " still exists" <<
            info << "remove " << a << " manually then re-run pkg-purge";
      }
    }
    else
    {
      assert (!p->out_root);
      pkg_purge_fs (c, t, p, !o.keep ());
    }

    // Finally, update the database state.
    //
    if (o.keep ())
    {
      if (p->state != package_state::fetched) // No-op we were talking about.
      {
        p->state = package_state::fetched;
        db.update (p);
        t.commit ();
      }
    }
    else
    {
      db.erase (p);
      t.commit ();
      p->state = package_state::transient;
    }

    if (verb)
        text << (o.keep () ? "keeping archive " : "purged ") << *p;

    return 0;
  }
}
